/**
 * WM TYPEWRITER SCROLLING
 */
"use strict";
(function(mod) {
	if (typeof exports == "object" && typeof module == "object") {
		mod(require("codemirror"));
	} else if (typeof define == "function" && define.amd) {
		define(["codemirror"], mod);
	} else {
		mod(CodeMirror);
	}
})(function(CodeMirror) {

	var oldPos = 0;
	var oldPosX = 0;
	var oldLinePos = 0;

	CodeMirror.defineOption("typewriterScrolling", false, function(cm, val, old) {
		if (old && old != CodeMirror.Init) {
			cm.off("changes", onChanges);
		}
		if (val) {
			cm.on("changes", onChanges);
		}
	});

	"use strict";
	CodeMirror.commands.scrollSelectionToCenter = function (cm) {
		if (cm.getOption("disableInput")) {
			return CodeMirror.Pass;
		}
		cm.scrollTo(null, getScrollPosition(cm).scrollTo);
		cm.refresh();
	};

	CodeMirror.commands.lockAtCurrentPosition = function (cm) {
		var top = cm.cursorCoords(true, "window").top;
		var max = cm.getWrapperElement().offsetHeight;
		localStorage["VIEWPORT_TYP_SCROLLING_POSITION"] = (max / top).toFixed(2);
	};



	function onChanges(cm, changes) {

		if (cm.getSelection().length !== 0) {
			return;
		}

		var currentPos = getScrollPosition(cm).currentPos;
		var posX = cm.cursorCoords(true, "local").left;
		var cursor = cm.getCursor();

		if ((currentPos !== oldPos && posX <= oldPosX) || Math.abs(cursor.line - oldLinePos) >= 2) {
			oldPos = currentPos;
			cm.execCommand("scrollSelectionToCenter");
		}

		oldPosX = posX;
		oldLinePos = cursor.line;

	}


	function getScrollPosition(cm) {
		var cursor = cm.getCursor("from");
		var top = cm.charCoords({line: cursor.line, ch: cursor.ch}, "local").bottom;

		var factor = localStorage["VIEWPORT_TYP_SCROLLING_POSITION"];
		factor = (factor) ? factor : 3;
		var halfWindowHeight = cm.getWrapperElement().offsetHeight / factor;

		var scrollTo = Math.round(top - halfWindowHeight);
		return {scrollTo: scrollTo, currentPos: top};
	}

});