/*
 ######    ##     ##       ########   ##     ##  ########
##    ##   ###   ###       ##          ##   ##      ##
##         #### ####       ##           ## ##       ##
##         ## ### ##       ######        ###        ##
##         ##     ##       ##           ## ##       ##
##    ##   ##     ##       ##          ##   ##      ##
 ######    ##     ##       ########   ##     ##     ##
*/

(function (mod) {
	if (typeof exports == "object" && typeof module == "object") // CommonJS
		mod(require("../../lib/codemirror"));
	else if (typeof define == "function" && define.amd) // AMD
		define(["../../lib/codemirror"], mod);
	else // Plain browser env
		mod(CodeMirror);
})(function (CodeMirror) {


	// Helper - get text from selections or lines under cursor
	CodeMirror.defineExtension("getSelectedOrLines", function (remove) {
		remove = (remove === "undefined") ? false : remove;

		if (this.somethingSelected()) {
			var txt = this.getSelections().join("\n");
			if (remove) this.replaceSelection("");
			return txt;
		} else { // nothing selected

			var selectionsA = this.listSelections();
			var str = "";
			for (var i = 0; i < selectionsA.length; i++) {
				str += this.getLine(selectionsA[i].from().line) + "\n\n";
				if (remove) this.execCommand("deleteLine");
			}
			return str;
		}
	});


	// Define some commands
	CodeMirror.commands.save = function (cm) {

		// fire event
		var saveE = new CustomEvent("saveE", {
			detail: {
				id: storage.getCurrentId(),
			}
		});
		document.body.dispatchEvent(saveE);
		storage.updateDocument(storage.getCurrentId());

		// visual manual save indicator
		$('#info_bar_name').fadeOut(300).fadeIn(150).fadeOut(150).fadeIn(400);

	};



	CodeMirror.commands.newDocument = function (cm) {
		storage.newDocument(storage.getCurrentId());
	};

	CodeMirror.commands.newDocumentFromFile = function (cm) {
		files.openWmFileWithDialog();
	};


	CodeMirror.commands.showCmdPalette = function (cm) {
		commandPalette.showCommandPalette();
	};



	CodeMirror.commands.copy = function (cm) {
		var clipboard = gui.Clipboard.get();
		clipboard.set(cm.getSelectedOrLines(false), 'text');
	};


	CodeMirror.commands.cut = function (cm) {
		var clipboard = gui.Clipboard.get();
		clipboard.set(cm.getSelectedOrLines(true), 'text');
	};


	CodeMirror.commands.copyAll = function (cm) {
		var clipboard = gui.Clipboard.get();
		clipboard.set(cm.getValue(), 'text');
	};


	// only of nothing is selected
	CodeMirror.commands.copySentence = function (cm) {

	};

	// only if nothing is selected
	CodeMirror.commands.cutSentence = function (cm) {

	};


	CodeMirror.commands.paste = function (cm) {
		var clipboard = gui.Clipboard.get();
		cm.replaceSelection(clipboard.get('text'));
	};


	CodeMirror.commands.pasteAndCollapse = function (cm) {
		var clipboard = gui.Clipboard.get();
		cm.operation(function () {
			cm.replaceSelection(clipboard.get('text'), "around");
			cm.execCommand("collapseSelection");
		});
	};



	CodeMirror.commands.scrollCursorToTypewriterPos = function (cm) {

		var top = cm.charCoords({
			line: cm.getCursor().line,
			ch: 0
		}, "local").top;
		var factor = localStorage["VIEWPORT_TYP_SCROLLING_POSITION"];
		factor = (factor) ? factor : 3;
		var halfWindowHeight = cm.getWrapperElement().offsetHeight / factor;
		cm.scrollTo(null, Math.round(top - halfWindowHeight));
	};

	CodeMirror.commands.toggleTypewriterScrolling = function (cm) {
		localStorage["VIEW_TYPEWRITER_SCROLLING"] = localStorage.getItem("VIEW_TYPEWRITER_SCROLLING") === null || localStorage["VIEW_TYPEWRITER_SCROLLING"] == 1 ? 0 : 1;
		cm.setOption("typewriterScrolling", !cm.getOption("typewriterScrolling"));
		wm.showNotice(cm.getOption("typewriterScrolling") ? "Typewriter scrolling ON" : "Typewriter scrolling OFF");
	};


	CodeMirror.commands.hardWrapAtPosition = function (cm) {

		var html = 'Hard wrap at: <br><input autofocus style="font-size: 2em; width: 95%; outline: none; border: none; background-color: transparent;"></input> <div style="font-size: 1em; float: right;">Hit enter to confirm!</div>';

		cm.openDialog(html, function (value) {

			v = parseInt(value, 10);
			localStorage['EDIT_HARD_WRAP_COLUMN'] = v;

			var s = cm.listSelections();
			for (var i = 0; i < s.length; i++) {
				cm.wrapParagraphsInRange(s[i].from(), s[i].to(), {
					column: v,
					// paragraphStart: /\n{1,2}/,
					// paragraphEnd: /\n{1,2}/
					// wrapOn: new RegExp('.\s')
				});
			}


		}, {
			value: localStorage['EDIT_HARD_WRAP_COLUMN'] ? localStorage['EDIT_HARD_WRAP_COLUMN'] : 80,
			selectValueOnOpen: true,
			closeOnEnter: true,
			closeOnBlur: true,
			onClose: function () {}
		});
	};

	CodeMirror.commands.removeNewLinesFromSelection = function (cm) {
		var sels = cm.listSelections();
		for (var i = 0; i < sels.length; i++) {
			var n = cm.getRange(sels[i].from(), sels[i].to()).replace(/\n\n/g, "[+*+]").replace(/\n/g, " ").replace(/\[\+\*\+\]/g, "\n\n");
			cm.replaceRange(n, sels[i].from(), sels[i].to());
		}
	};


	// CodeMirror.commands.toggleSyntaxHighlighting = function(cm) {

	// 	var text = cm.getValue();
	// 	var cursor = cm.getCursor();

	// 	var mode = cm.getOption("mode");
	// 	var backdrop = cm.getOption("backdrop");

	// 	var isSH = (mode == "gfm" || backdrop == "gfm");
	// 	var isSpell = mode == "spell-checker";


	// 	if (isSH) { // disable
	// 		cm.execCommand("unfoldAll");
	// 		cm.setOption("highlightComments", false);

	// 		cm.setOption("mode", isSpell ? "spell-checker" : "none");
	// 		cm.setOption("backdrop", "none");
	// 	} else { // enable

	// 		cm.setOption("highlightComments", true);

	// 		cm.setOption("mode", isSpell ? "spell-checker" : "gfm");
	// 		cm.setOption("backdrop", isSpell ? "gfm" : "none");
	// 	}


	// 	cm.setValue(text);
	// 	cm.setCursor(cursor);
	// 	cm.execCommand("scrollCursorToTypewriterPos");

	// 	cm.refresh();
	// 	wm.showNotice(isSH ? "Syntax highlighting OFF" : "Syntax highlighting ON");
	// };



	CodeMirror.commands.expandSnippet = function (cm) {
		repository.setReplacementFromTrigger();
	};


	CodeMirror.commands.toggleStyleActiveLine = function (cm) {
		if (localStorage["VIEW_STYLE_ACTIVE_LINE"] && localStorage["VIEW_STYLE_ACTIVE_LINE"] == "1") {
			localStorage["VIEW_STYLE_ACTIVE_LINE"] = "0";
			cm.setOption("styleActiveLine", false);
		} else {
			localStorage["VIEW_STYLE_ACTIVE_LINE"] = "1";
			cm.setOption("styleActiveLine", true);
		}
	};



	/*
	##    ##   ########   ##    ##   ##     ##     ###     ########
	##   ##    ##          ##  ##    ###   ###    ## ##    ##     ##
	##  ##     ##           ####     #### ####   ##   ##   ##     ##
	#####      ######        ##      ## ### ##  ##     ##  ########
	##  ##     ##            ##      ##     ##  #########  ##
	##   ##    ##            ##      ##     ##  ##     ##  ##
	##    ##   ########      ##      ##     ##  ##     ##  ##
	*/



	CodeMirror.defineInitHook(function (cm) {
		var isMac = (os.platform() === 'darwin');
		var ctrl = (isMac) ? "Cmd-" : "Ctrl-";

		// create map
		var map = CodeMirror.keyMap.writemonkey = {
			fallthrough: "default"
		};


		map['Shift-' + ctrl + 'C'] = function (cm) {
			cm.execCommand("copyAll");
		};

		map['Shift-' + ctrl + 'V'] = function (cm) {
			cm.execCommand("pasteAndCollapse");
		};

		map[ctrl + 'N'] = function (cm) {
			cm.execCommand("newDocument");
		};

		map['Shift-' + ctrl + 'N'] = function (cm) {
			cm.execCommand("newDocumentFromFile");
		};

		map[ctrl + 'B'] = function (cm) {
			cm.execCommand("toggleBold");
		};

		map[ctrl + 'I'] = function (cm) {
			cm.execCommand("toggleItalic");
		};

		map[ctrl + 'O'] = function (cm) {
			cm.execCommand("toggleCode");
		};

		map[ctrl + 'T'] = function (cm) {
			cm.execCommand("toggleStrikethrough");
		};

		map[ctrl + 'R'] = function (cm) {
			cm.execCommand("toggleComment");
		};

		map[ctrl + 'E'] = function (cm) {
			cm.execCommand("toggleMarked");
		};

		map[ctrl + '0'] = function (cm) {
			cm.execCommand("resetZoom");
		};

		map[ctrl + '6'] = function (cm) {
			cm.execCommand("decreaseFontZoom");
		};

		map[ctrl + '7'] = function (cm) {
			cm.execCommand("increaseFontZoom");
		};

		map[ctrl + '4'] = function (cm) {
			cm.execCommand("decreaseTextAreaWidth");
		};

		map[ctrl + '5'] = function (cm) {
			cm.execCommand("increaseTextAreaWidth");
		};

		map['Shift-' + ctrl + 'T'] = function (cm) {
			cm.execCommand("toggleTypewriterScrolling");
		};

		map['Shift-' + ctrl + 'M'] = function (cm) {
			cm.execCommand("moveToRepository");
		};

		map['Shift-' + ctrl + 'L'] = function (cm) {
			cm.execCommand("copyToRepository");
		};

		map['Shift-' + ctrl + 'E'] = function (cm) {
			cm.execCommand("expandSnippet");
		};

		map[ctrl + 'F2'] = function (cm) {
			cm.execCommand("newWmBookmark");
		};

		map['Shift-' + ctrl + 'F2'] = function (cm) {
			cm.execCommand("clearWmBookmarks");
		};

		map['F2'] = function (cm) {
			cm.execCommand("nextWmBookmark");
		};

		map['Shift-F2'] = function (cm) {
			cm.execCommand("prevWmBookmark");
		};


		map['F4'] = function (cm) {
			cm.execCommand("jumpToLastChangeSpot");
		};

		map['F5'] = function (cm) {
			cm.execCommand("collapseSelection");
		};

		map['F6'] = function (cm) {
			cm.execCommand("toggleFocus");
		};

		map[ctrl + 'F6'] = function (cm) {
			cm.execCommand("toggleFadeInactiveParagraphs");
		};

		map[ctrl + 'F7'] = function (cm) {
			cm.execCommand("toggleStyleCurrentSentence");
		};

		map[ctrl + 'F8'] = function (cm) {
			cm.execCommand("toggleSentenceBreakup");
		};

		map[ctrl + 'F9'] = function (cm) {
			cm.execCommand("toggleAnchoredReading");
		};

		map['F7'] = function (cm) {
			cm.execCommand("toggleSpelling");
		};

		map['F8'] = function (cm) {
			cm.execCommand("toggleWhiteSpaces");
		};

		map[ctrl + '9'] = function (cm) {
			cm.execCommand("toggleFoldAll");
		};

		map[ctrl + '8'] = function (cm) {
			cm.execCommand("toggleFold");
		};

		map[ctrl + '1'] = function (cm) {
			cm.execCommand("showSelectionMatches");
		};

		map[ctrl + 'D'] = function (cm) {
			cm.execCommand("selectNextOccurrence");
		};

		map['Home'] = function (cm) {
			cm.execCommand("goLineLeft");
		};

		map['End'] = function (cm) {
			cm.execCommand("goLineRight");
		};

		map["Shift-" + 'Home'] = function (cm) {
			cm.setExtending(true);
			cm.execCommand("goLineLeft");
			cm.setExtending(false);
		};

		map["Shift-" + 'End'] = function (cm) {
			cm.setExtending(true);
			cm.execCommand("goLineRight");
			cm.setExtending(false);
		};


		// disable sorting
		// map[ctrl + 'F9'] = function (cm) {
		// };
		
		map['F9'] = function (cm) {};


		// util for markdown list
		map['Enter'] = function (cm) {
			cm.execCommand("newlineAndIndentContinueMarkdownList");
		};


		map['Tab'] = function (cm) {
			var selections = cm.listSelections();
			cm.operation(function () {
				for (var i = 0; i < selections.length; i++) {
					var line = selections[i].head.line;
					var lineText = cm.getLine(line);

					if (/^\s*[-*+]\s/.test(lineText)) {
						cm.indentLine(line, "add");
					} else {
						var rpl = (cm.getOption('indentWithTabs')) ? '\t' : ' '.repeat(cm.getOption('tabSize'));
						cm.replaceSelection(rpl);
					}
				}
			});
		};

		map['Shift-Tab'] = function (cm) {
			var selections = cm.listSelections();
			cm.operation(function() {
				for (var i = 0; i < selections.length; i++) {
					var line = selections[i].head.line;
					cm.indentLine(line, "subtract");
				}
			});
		};


		// set markdown heading level up
		map['Alt-Up'] = function (cm) {
			// headings up
			cm.execCommand("headingsUp");


		};
		// set markdown heading level down
		map['Alt-Down'] = function (cm) {
			// headings down
			cm.execCommand("headingsDown");

		};


		// clean text
		map[ctrl + "K " + ctrl + "C"] = function (cm) {
			cm.execCommand("cleanText");
		};

		// csearch in google
		map[ctrl + "K " + ctrl + "G"] = function (cm) {
			cm.execCommand("searchWithGoogle");
		};

		// add key map
		cm.addKeyMap(map, false);

	});



	CodeMirror.defineExtension("getMarks", function () {
		var marks = this.getAllMarks();

		var folds = [];
		var bookmarks = [];
		var collapsed = [];

		if (marks.length) {
			var le = marks.length;
			for (var i = 0; i < le; i++) {
				if (marks[i].__isFold && marks[i].collapsed) {
					var pos = marks[i].find().from;
					var p = {
						line: pos.line,
						ch: pos.ch
					};
					folds.push(p);
				} else if (marks[i].type == "bookmark") {
					var pos = marks[i].find();
					bookmarks.push(pos);
				} else if (marks[i].type == "collapsed_region") {
					var pos = marks[i].find();
					collapsed.push(pos);
				}
			}
		}

		return {
			folds: folds,
			bookmarks: bookmarks,
			collapsed: collapsed
		};
	});

	CodeMirror.defineExtension("setFolds", function (folds) {
		var le = folds.length;
		if (le <= 0) {
			return;
		} else {
			for (var i = le - 1; i >= 0; i--) {
				if (!this.isFolded(folds[i])) {
					this.foldCode(folds[i]);
				}
			}
		}
	});


	/*
	########    #######    ######    ##     ##   ######
	##         ##     ##  ##    ##   ##     ##  ##    ##
	##         ##     ##  ##         ##     ##  ##
	######     ##     ##  ##         ##     ##   ######
	##         ##     ##  ##         ##     ##        ##
	##         ##     ##  ##    ##   ##     ##  ##    ##
	##          #######    ######     #######    ######
	*/


	CodeMirror.commands.toggleFocus = function (cm) {
		cm.setFocus();
	}


	// CodeMirror.defineExtension("isFocusOn", function() {
	// 	var mks = this.getAllMarks();

	// 	for (var i = 0; i < mks.length; i++) {
	// 		if (mks[i].type == "focus_region") {
	// 			return true;
	// 		}
	// 	}
	// 	return false;
	// });



	CodeMirror.defineExtension("setFocus", function () {
		var cm = this;

		if (cm.getFocusedBlock()) { // exit focus

			$("#segment-focus-icon").remove();

			var allMarks = cm.getAllMarks();

			//cm.execCommand("selectAll");
			//cm.execCommand("collapseSelection");

			for (var i = 0; i < allMarks.length; i++) {
				if (allMarks[i].type == "focus_region") {
					allMarks[i].clear();
				}
			}

			// wm.showNotice("Focus off");

			// fire event
			var focusE = new CustomEvent("focusE", {
				detail: {
					isExit: true
				}
			});
			document.body.dispatchEvent(focusE);

		} else { // enable focus

			// set focus
			var cur = cm.getCursor();
			var last = cm.lastLine();

			var aboveFrom = {
				line: 0,
				ch: 0
			};
			var belowTo = {
				line: last,
				ch: cm.getLine(last).length
			};
			var aboveTo, belowFrom;

			var header = cm.getHeaderLevel(cur);

			if (cm.somethingSelected()) { // selection
				var selections = cm.listSelections();
				aboveTo = {
					line: selections[0].from().line,
					ch: selections[0].from().ch
				};
				belowFrom = {
					line: selections[0].to().line,
					ch: selections[0].to().ch
				};
				// wm.showNotice("Focus on selection");
			} else if (header != -1) { // heading
				var lineToCheck = cur.line;
				var headerEndLine = false;
				while (!headerEndLine) {
					lineToCheck++;
					var nextLine = cm.getHeaderLevel({
						line: lineToCheck,
						ch: 0
					});
					if (nextLine != -1 && nextLine <= header) {
						headerEndLine = true;
						aboveTo = {
							line: cur.line,
							ch: 0
						};
						belowFrom = {
							line: lineToCheck - 1,
							ch: cm.getLine(lineToCheck - 1).length
						};
					} else if (last == lineToCheck) {
						headerEndLine = true;
						aboveTo = {
							line: cur.line,
							ch: 0
						};
						belowFrom = {
							line: last,
							ch: cm.getLine(last).length
						};
					}
				}

			} else { // paragraph / sentence
				// aboveTo = {line:cur.line, ch:0};
				// belowFrom = {line:cur.line, ch:cm.getLine(cur.line).length};

				var sentence = cm.findSentenceAt(cm.getLine(cur.line), cur);
				aboveTo = {
					line: cur.line,
					ch: sentence.start
				};
				belowFrom = {
					line: cur.line,
					ch: sentence.end
				};

			}

			// console.log(aboveFrom, aboveTo, belowFrom, belowTo)



			// check if setting markers is necessary
			if (aboveTo.line == 0 && aboveTo.ch == 0 && belowFrom.line == last && belowFrom.ch == cm.getLine(last).length) {
				return false;
			}




			var markerOption = {
				collapsed: true,
				inclusiveLeft: true,
				// inclusiveRight: false,
				addToHistory: true,
				readOnly: true,
				type: "focus_region"
			};

			var markerOption2 = {
				collapsed: true,
				inclusiveRight: true,
				// inclusiveLeft: false,
				addToHistory: true,
				readOnly: true,
				type: "focus_region"
			};

			try {
				// set markers
				var mk1 = cm.markText(aboveFrom, aboveTo, markerOption);
				var mk2 = cm.markText(belowFrom, belowTo, markerOption2);

				// fire event
				var focusE = new CustomEvent("focusE", {
					detail: {
						isExit: false
					}
				});
				document.body.dispatchEvent(focusE);

				// set focus icon
				var focusIcon = document.createElement("div");
				$(focusIcon)
					.attr("id", "segment-focus-icon")
					.css({
						// "width": "15px",
						"margin-right": "auto",
						"margin-left": "auto",
						"font-family": "Font Awesome",
						"font-size": ".8em",
						"line-height": "1em",
						"text-align": "center",
						"cursor": "pointer",
						"opacity": "1"
					})
					// .html("exit focus");
					.html("&#xf052;"); // &#xf08b; &#xf0dc; &#xf06e;

				$("#info_bar_top_middle").prepend(focusIcon);

				$(focusIcon).on("click", function (e) {
					e.stopPropagation();
					_editor.focus();
					_editor.setFocus();
				});

			} catch (e) {
				wm.showNotice("Can't focus on sentence!");
			}
		}

		//cm.execCommand("scrollCursorToTypewriterPos");
		cm.scrollIntoView(null, 300);
		// cm.focus();

	});

	CodeMirror.defineExtension("getFocusedBlock", function () { // returns false if no focus

		var allMarks = this.getAllMarks();
		var focusMarks = [];
		for (var i = 0; i < allMarks.length; i++) {
			if (allMarks[i].type == "focus_region") {
				focusMarks.push(allMarks[i].find());
			}
		}
		if (focusMarks.length < 1) return false;

		var start, stop;
		if (focusMarks.length == 2) {
			start = focusMarks[0].to;
			stop = focusMarks[1].from;
		} else if (focusMarks.length == 1) {
			if (focusMarks[0].from.line == 0 && focusMarks[0].from.ch == 0) { // at the end
				start = {
					line: focusMarks[0].to.line,
					ch: focusMarks[0].to.ch
				};
				stop = {
					line: this.lastLine(),
					ch: this.getLine(this.lastLine()).length - 1
				};
			} else { // at the top
				start = {
					line: 0,
					ch: 0
				};
				stop = {
					line: focusMarks[0].from.line,
					ch: focusMarks[0].from.ch
				};
			}
		}
		return {
			start: start,
			stop: stop,
			text: this.getRange(start, stop)
		};
	});



	CodeMirror.defineExtension("getHeaderLevel", function (cursor) { // -1 if no header
		var h = this.getTokenTypeAt(cursor);
		if (h) {
			var m = h.match(/\bheader-([0-9])\b/);
			if (m && m[1]) {
				return m[1];
			} else {
				return -1;
			}
		} else {
			return -1;
		}
	});


	/*
	 ######    ##         ########    ######    ########
	##    ##   ##         ##     ##  ##    ##   ##     ##
	##         ##         ##     ##  ##         ##     ##
	##         ##         ########    ######    ##     ##
	##         ##         ##               ##   ##     ##
	##    ##   ##         ##         ##    ##   ##     ##
	 ######    ########   ##          ######    ########
	*/


	CodeMirror.commands.toggleFoldAll = function (cm) {

		var marks = cm.getMarks();
		if (marks.folds.length >= 1) {
			cm.execCommand("unfoldAll");
		} else {
			cm.execCommand("foldAll");
		}
	};




	CodeMirror.commands.collapseSelection = function (cm) {
		// check if markers is at cursor
		var collapsedMark = null;
		var presentMarks = cm.findMarksAt(cm.getCursor());
		for (var i = 0; i < presentMarks.length; i++) {
			if (presentMarks[i].type == "collapsed_region") {
				collapsedMark = presentMarks[i];
				break;
			}
		}

		// if (presentMarks.some(function(val, index, arr) {return val.type == "collapsed_region"}))
		// if (presentMarks.length && presentMarks[0].type == "collapsed_region")

		if (collapsedMark != null) { // remove marker
			// cm.setSelection(collapsedMark.find().from, collapsedMark.find().to);
			var from = collapsedMark.find().from;
			var to = collapsedMark.find().to;
			collapsedMark.clear();
			cm.setSelection(from, to);


		} else { // set marker
			var selections = cm.listSelections();
			if (cm.somethingSelected()) { // only act on selection
				for (var i = 0; i < selections.length; i++) {
					cm.collapseRegion(selections[i].from(), selections[i].to());
				}
			} else { // nothing is selected, act on cursor
				for (var i = 0; i < selections.length; i++) {
					var lineNum = selections[i].from().line;

					var from = {
						line: lineNum,
						ch: 0
					};
					var to = {
						line: lineNum,
						ch: cm.getLine(lineNum).length
					};

					cm.collapseRegion(from, to);
				}
			}
		}
	};


	CodeMirror.commands.expandAllRegions = function (cm) {
		var marks = cm.getAllMarks();
		for (var i = 0; i < marks.length; i++) {
			if (marks[i].type == "collapsed_region") {
				marks[i].clear();
			}
		}
		wm.showNotice("All regions expanded");
	};


	CodeMirror.defineExtension("setCollapsed", function (regions) {
		var le = regions.length;
		if (le <= 0) {
			return;
		} else {
			for (var i = le - 1; i >= 0; i--) {
				this.collapseRegion(regions[i].from, regions[i].to);
			}
		}
	});



	CodeMirror.defineExtension("collapseRegion", function (from, to) {
		var rangeText = this.getRange(from, to);


		var rangeWordCount = stats.getWords(rangeText).count;
		var rangeTitle = '<span class="awf" style="opacity: .5;">&#xf066;&nbsp;&nbsp;</span>' + rangeText.trim().replace(/\n/gi, " ").substring(0, _settings.collapsed_region_placeholder_length).trim();


		var widget = document.createElement("span");
		var widgetCSS = {};
		var widgetHTML = "";


		// is image
		var isImage = getImageSecFromMd(rangeText);
		var isTable = getMdTable(rangeText);
		var isKatex = getKatex(rangeText);
		var isComment = getComment(rangeText);

		if (isImage != -1) {
			widgetCSS = {
				"display": "inline-block",
				"width": _settings.collapsed_region_image_size[0],
				"height": _settings.collapsed_region_image_size[1],
				"margin": ".5em .5em .5em .5em",
				"background-image": 'url(' + isImage + ')',
				"background-position": "center top",
				"background-color": "transparent",
				"background-size": "cover",
				"background-repeat": "no-repeat",
				"border-radius": "6px",
				"cursor": "pointer"
			}
		} else if (isTable != -1) {
			widgetCSS = {
				"display": "inline-block",
				"background-color": "transparent",
				"color": "inherit",
				"max-width": "95%",
				"cursor": "pointer"
			};
			widgetHTML = '<div class="collapsed-inline-table">' + mdIt.render(isTable) + '</div>';
		} else if (isKatex != -1) {
			widgetCSS = {
				"display": "inline-block",
				"align-content": "center",
				"background-color": "transparent",
				"padding": "0em 0em 0em 0em",
				"margin": "0em 0em 0em 0em",
				"color": "inherit",
				"max-width": "95%",
				"cursor": "pointer"
			};
			widgetHTML = '<span class="katex">' + isKatex + '</span>';

		} else if (isComment != -1) {
			var randRotate = (Math.random() * _settings.collapsed_region_comment_max_tilt) - _settings.collapsed_region_comment_max_tilt / 2;

			widgetCSS = {
				"display": "inline-block",
				"font-family": _settings.collapsed_region_comment_font,
				"font-size": _settings.collapsed_region_comment_font_size,
				"align-content": "center",
				"padding": ".5em 1.5em",
				"max-width": _settings.collapsed_region_comment_max_width,
				"margin": ".5em .5em .5em .5em",
				"cursor": "pointer",
				"filter": "invert(100%)",
				"transform": 'rotate(' + randRotate + 'deg)'
			};
			widgetHTML = '<div class="collapsed-inline-comment">' + mdIt.render(isComment) + '</div>';
			// theme-background-color-mark theme-text-color-inverse
		} else {
			var countLabel = document.createElement("span");
			$(countLabel).css({
					"position": "relative",
					"font-family": "Arial, sans-serif",
					"top": "-.1em",
					"border-radius": "2em",
					"padding": ".3em .6em .3em .6em",
					"margin-left": "1em",
					"margin-right": ".3em",
					"font-size": ".65em",
					"min-width": ".5em",
					"text-align": "center"
					// "color": "black",
					// "background-color": "#bcbcbc",
					// "opacity": ".5"
				}).addClass('theme-background-color theme-text-color')
				.html(rangeWordCount + " wrd");

			$(countLabel).jqxTooltip({
				theme: 'metro',
				autoHide: false,
				showArrow: false,
				animationShowDelay: 50,
				animationHideDelay: 1000,
				autoHideDelay: 3000,
				opacity: 1,
				content: '<div style="font-size: 1.1em!important; max-height: 70vh; overflow: hidden; text-overflow: ellipsis; word-wrap: break-word; width: auto; padding: 1em; text-align: left;">' + rangeText.replace(/\n/g, '<br>').substring(0, 3000) + '</div><div>&nbsp;</div>',
				position: 'mouse',
				name: 'hidden_text',
				trigger: 'hover',
				width: 400,
				closeOnClick: true
			});

			widgetHTML = rangeTitle + "<span>&#8230;</span>";

		}

		$(widget)
			.css(widgetCSS)
			.addClass("collapsed-placeholder")
			.html(widgetHTML).append(countLabel);



		var marker = this.markText(from, to, {
			atomic: true,
			replacedWith: widget,
			handleMouseEvents: true,
			addToHistory: true,
			readOnly: false,
			inclusiveRight: false,
			inclusiveLeft: false,
			title: rangeTitle,
			type: "collapsed_region"
		});

		this.setCursor(to.line, to.ch);


		widget.addEventListener('click', function (e) {

			_editor.operation(function () {
				var startPos = marker.find().from;
				var endPos = marker.find().to;
				marker.clear();
				_editor.setSelection(startPos, endPos);


				// check if ctrl is pressed -- enter focus
				if (e.metaKey || e.ctrlKey) {
					_editor.setFocus();
					_editor.setSelection(startPos, startPos);
				}
			});

		}, false);

		if (isKatex != -1) {

			var disp = false;
			isKatex = isKatex.trim();
			if ((isKatex.startsWith("$$$") && isKatex.endsWith("$$$"))) {
				disp = true;
			}

			katex.render(isKatex.replace(/^\${2,3}/g, "").replace(/\${2,3}$/g, ""), widget, {
				throwOnError: false,
				displayMode: disp,
				strict: false,
				errorColor: "#8b0000"
			});
		}
	});


	// drop

	CodeMirror.defineInitHook(function (cm) {
		cm.on("drop", function (cm, e) {
			if (e.dataTransfer.files.length > 0) {
				if (e.dataTransfer.files[0].type == "text/plain" ||
					e.dataTransfer.files[0].type == "text/markdown"
				) {
					e.codemirrorIgnore = true;
					e.preventDefault();
					files.openWmFile(e.dataTransfer.files[0].path);
				} else if (e.dataTransfer.files[0].type == "image/jpeg" ||
					e.dataTransfer.files[0].type == "image/png" ||
					e.dataTransfer.files[0].type == "image/gif" ||
					e.dataTransfer.files[0].type == "image/webp" ||
					e.dataTransfer.files[0].type == "image/tiff"
				) {
					e.codemirrorIgnore = true;
					e.preventDefault();
					var path = encodeURI(e.dataTransfer.files[0].path.replace(/\\/g, "/"));
					cm.replaceSelection("![](" + path + ")");
					cm.execCommand("collapseSelection");
				}
			}
		});
	});





	/*
	########    #######    #######   ##    ##   ##     ##
	##     ##  ##     ##  ##     ##  ##   ##    ###   ###
	##     ##  ##     ##  ##     ##  ##  ##     #### ####
	########   ##     ##  ##     ##  #####      ## ### ##
	##     ##  ##     ##  ##     ##  ##  ##     ##     ##
	##     ##  ##     ##  ##     ##  ##   ##    ##     ##
	########    #######    #######   ##    ##   ##     ##
	*/


	// Jump to the lat change spot

	CodeMirror.defineInitHook(function (cm) {
		cm.on("change", function (cm, ch) {
			_lastChangePosition = cm.getCursor();
		});
	});

	CodeMirror.commands.jumpToLastChangeSpot = function (cm) {

		if (_lastChangePosition !== null) {
			cm.setCursor(_lastChangePosition, {
				scroll: false
			});
			cm.execCommand("scrollCursorToTypewriterPos");
			cm.execCommand("showScrollBar");
		}
	};


	CodeMirror.commands.newWmBookmark = function (cm) {
		cm.insertWmBookmarkAtPos(cm.getCursor());
	};

	CodeMirror.commands.nextWmBookmark = function (cm) {
		cm.goNextWmBookmark(false);
	};

	CodeMirror.commands.prevWmBookmark = function (cm) {
		cm.goNextWmBookmark(true);
	};

	CodeMirror.commands.clearWmBookmarks = function (cm) {
		var bkm = cm.getAllMarks();
		for (var i = 0; i < bkm.length; i++) {
			if (bkm[i].type == "bookmark")
				bkm[i].clear();
		}
		// cm.bookmarksPositions = [];
		// cm.updateBookmarksOnScrollbar();
		wm.showNotice("Bookmarks cleared");
	};


	CodeMirror.defineExtension("setBookmarks", function (bookmarks) {

		// reset bookmarks on scrollbar
		// if (this.scrlbAnnot) this.scrlbAnnot.clear();
		// this.scrlbAnnot = this.annotateScrollbar({
		// 	listenForChanges: false,
		// 	className: "bookmark-on-scrollbar theme-background-color-mark"
		// });
		// this.bookmarksPositions = [];


		var le = bookmarks.length;
		if (le <= 0) {
			return;
		} else {
			for (var i = le - 1; i >= 0; i--) {
				this.insertWmBookmarkAtPos(bookmarks[i]);
			}
			// this.updateBookmarksOnScrollbar();
		}
	});


	CodeMirror.defineExtension("goNextWmBookmark", function (isUp) {

		// check if no bookmarks - dirty way
		var mrksCheck = this.findMarks({
			line: 0,
			ch: 0
		}, {
			line: this.lastLine(),
			ch: this.getLine(this.lastLine()).length
		});
		var bkmsCheck = [];
		for (var i = 0; i < mrksCheck.length; i++) {
			if (mrksCheck[i].type == "bookmark") {
				bkmsCheck.push(mrksCheck[i]);
			}
		}
		if (bkmsCheck.length <= 0) {
			wm.showNotice("No bookmarks in document!")
			return;
		}



		var start = (isUp) ? {
			line: 0,
			ch: 0
		} : this.getCursor();
		var end = (isUp) ? this.getCursor() : {
			line: this.lastLine(),
			ch: this.getLine(this.lastLine()).length
		};

		var mrks = this.findMarks(start, end);


		var bkms = [];
		for (var i = 0; i < mrks.length; i++) {
			if (mrks[i].type == "bookmark") {
				bkms.push(mrks[i]);
			}
		}

		if (bkms.length <= 0) { // no more bookmarks in given direction
			if (isUp) {
				this.setCursor({
					line: this.lastLine(),
					ch: this.getLine(this.lastLine()).length
				}, {
					scroll: false
				});
				this.goNextWmBookmark(true);
			} else {
				this.setCursor({
					line: 0,
					ch: 0
				}, {
					scroll: false
				});
				this.goNextWmBookmark(false);
			}
			return;
		}

		var current = (isUp) ? bkms[bkms.length - 1].find() : bkms[0].find();

		if (current) {
			this.setCursor(current, {
				scroll: false
			});
			this.scrollIntoView(this.getCursor(), 200);
			this.execCommand("showScrollBar");
		}

	});


	CodeMirror.defineExtension("insertWmBookmarkAtPos", function (pos) {

		var bkmNode = document.createElement("span");
		$(bkmNode).css({
			// "width": "20px",
			// "height": "1em",
			// "border-radius": "1em",
			// "padding": "3px",
			"margin-left": "5px",
			"margin-right": "5px",
			// "display": "inline",
			"font-size": ".7em",
			"font-family": "Font Awesome",
			// "color": "red",
			// "background-color": "red",
			"cursor": "pointer"
		}).addClass("theme-mark-color");
		$(bkmNode).html("&#xf02e;");


		var bkm = this.setBookmark(pos, {
			insertLeft: true,
			widget: bkmNode,
			handleMouseEvents: false
		});

		//for scrollbar annotations
		// this.bookmarksPositions.push({from:pos, to:pos});


		// event for removing bookmark
		bkmNode.addEventListener('click', function (e) {
			bkm.clear();
			// this.updateBookmarksOnScrollbar();
			_editor.focus();
		}, false);

	});



	// CodeMirror.defineExtension("updateBookmarksOnScrollbar", function() {

	// 	// get all bookmarks in correct order

	// 	var bkm = this.getAllMarks();
	// 	for (var i = 0; i < bkm.length; i++) {
	// 		if (bkm[i].type == "bookmark")
	// 			this.bookmarksPositions.push({from: bkm[i].find(), to: bkm[i].find()});
	// 	}

	// 	this.scrlbAnnot.update(this.bookmarksPositions);
	// 	console.log(this.bookmarksPositions);
	// });


	/*
	########      ###      ######    ##    ##    ######
	   ##        ## ##    ##    ##   ##   ##    ##    ##
	   ##       ##   ##   ##         ##  ##     ##
	   ##      ##     ##   ######    #####       ######
	   ##      #########        ##   ##  ##           ##
	   ##      ##     ##  ##    ##   ##   ##    ##    ##
	   ##      ##     ##   ######    ##    ##    ######
	*/


	CodeMirror.defineInitHook(function (cm) {
		CodeMirror.on(cm.getWrapperElement(), "mousedown", function (e) {
			var token = cm.getTokenAt(cm.getCursor());

			var line = cm.getCursor().line;
			var lineText = cm.getLine(line);

			var from = {
				line: line,
				ch: 0
			};
			var to = {
				line: line,
				ch: lineText.length
			};

			if (token.type && token.type.indexOf("formatting-task property") > -1) {
				e.preventDefault();
				//cm.setSelection(start, end);
				cm.replaceRange(lineText.replace("[x]", "[ ]"), from, to);
				jumps.updateListBox();
			} else if (token.type && token.type.indexOf("formatting-task meta") > -1) {
				e.preventDefault();
				//cm.setSelection(start, end);
				cm.replaceRange(lineText.replace("[ ]", "[x]"), from, to);
				jumps.updateListBox();
			}
		});
	});



	/*
	 ######     #######   ##     ##  ##     ##
	##    ##   ##     ##  ###   ###  ###   ###
	##         ##     ##  #### ####  #### ####
	##         ##     ##  ## ### ##  ## ### ##
	##         ##     ##  ##     ##  ##     ##
	##    ##   ##     ##  ##     ##  ##     ##
	 ######     #######   ##     ##  ##     ##
	*/
	// Defines overlay mode for wm comments
	CodeMirror.defineOption("highlightComments", false, function (cm, val, prev) {
		if (prev == CodeMirror.Init) prev = false;
		if (prev && !val)
			cm.removeOverlay("wm-comments");
		else if (!prev && val)

			cm.addOverlay({
				name: "wm-comments",
				token: function (stream, state) {
					var ch;

					if (stream.match("//") && stream.string.charAt(stream.pos - 3) !== ":") {

						while ((ch = stream.next()) != null) {

							if (ch === "/" && stream.next() === "/") {
								if (stream.string.charAt(stream.current().length - 3) !== ":") {
									return "wm-comments";
								} else {
									stream.next();
									stream.next();
								}
							} else if (stream.eol()) {
								return "wm-comments";
							}
						}
					} else {
						stream.next();
						return null;
					}

					// while (stream.next() != null && !stream.match("//", false)) {}
					// return null;
				}
			});
	});

	// Defines overlay mode for marked
	CodeMirror.defineOption("highlightMarked", false, function (cm, val, prev) {
		if (prev == CodeMirror.Init) prev = false;
		if (prev && !val)
			cm.removeOverlay("wm-marked");
		else if (!prev && val)

			cm.addOverlay({
				name: "wm-marked",
				token: function (stream, state) {
					var ch;

					if (stream.match(/^==[^=\t ]/)) {
						stream.backUp(2);
						while ((ch = stream.next()) != null) {
							if (stream.match(/^\S==/)) {

								return "wm-marked";
							} else if (stream.eol()) {
								return null;
							}
						}
					} else {
						stream.next();
						return null;
					}

				}
			});
	});


	// Defines overlay mode for footnotes
    CodeMirror.defineOption("highlightFootnotes", false, function (cm, val, prev) {
        if (prev == CodeMirror.Init) prev = false;
        if (prev && !val)
            cm.removeOverlay("wm-footnote");
        else if (!prev && val)

            cm.addOverlay({
                name: "wm-footnote",
                token: function (stream, state) {
                    var ch;

                    if (stream.match(/\^\[/) || stream.match(/\[\^\w+/)) {
                        stream.backUp(1);
                        while ((ch = stream.next()) != null) {
                            if (stream.match(/\]:?/)) {
                                return "wm-footnote";
                            } else if (stream.eol()) {
                                return null;
                            }
                        }
                    } else {
                        stream.next();
                        return null;
                    }
                }
            });
    });





	/*

	########  ########    ###    ########   #######  ##    ## ##       ##    ##
	##     ## ##         ## ##   ##     ## ##     ## ###   ## ##        ##  ##
	##     ## ##        ##   ##  ##     ## ##     ## ####  ## ##         ####
	########  ######   ##     ## ##     ## ##     ## ## ## ## ##          ##
	##   ##   ##       ######### ##     ## ##     ## ##  #### ##          ##
	##    ##  ##       ##     ## ##     ## ##     ## ##   ### ##          ##
	##     ## ######## ##     ## ########   #######  ##    ## ########    ##

	*/


	CodeMirror.defineExtension("enableReadOnly", function () {
		var lastLine = this.lastLine();
		var lastLineLength = this.getLine(lastLine).length;
		_isReadOnlyMarker = this.markText({
			line: 0,
			ch: 0
		}, {
			line: lastLine,
			ch: lastLineLength
		}, {
			className: "cm-document-is-read-only",
			inclusiveLeft: true,
			inclusiveRight: true,
			readOnly: true
		})

		// wm.showNotice("Document is marked as Read Only.<br><br>Editing is disabled.", {});
	});

	CodeMirror.defineExtension("disableReadOnly", function () {
		if (_isReadOnlyMarker !== null) {
			_isReadOnlyMarker.clear();
			_isReadOnlyMarker = null;

			// wm.showNotice("Read Only disabled.<br><br>Editing is allowed.", {});
		}
	});


	/*
	##      ## ########   ########   ########   ####  ##    ##   ########   ########
	##  ##  ## ##     ##  ##     ##  ##     ##   ##   ###   ##   ##     ##     ##
	##  ##  ## ##     ##  ##     ##  ##     ##   ##   ####  ##   ##     ##     ##
	##  ##  ## ########   ########   ##     ##   ##   ## ## ##   ##     ##     ##
	##  ##  ## ##   ##    ##         ##     ##   ##   ##  ####   ##     ##     ##
	##  ##  ## ##    ##   ##         ##     ##   ##   ##   ###   ##     ##     ##
	 ###  ###  ##     ##  ##         ########   ####  ##    ##   ########      ##
	*/

	// Needed for this to work
	// .CodeMirror pre > * { text-indent: 0px; }

	CodeMirror.defineInitHook(function (cm) {

		cm.on("renderLine", function (cm, line, elt) {
			if (cm.getOption("mode") == "gfm" || cm.getOption("backdrop") == "gfm") {
				var charWidth = cm.defaultCharWidth();
				var basePadding = 4;
				var leadingSpaceListBulletsQuotes = /^\s*([*+-]\s+|\d+\.\s+|>\s*)*/;
				// var leadingSpaceListBulletsQuotes = /^([*+-]\s+|\d+\.\s+|>\s*)*/; // without leading space
				var leading = (leadingSpaceListBulletsQuotes.exec(line.text) || [""])[0];
				var off = CodeMirror.countColumn(leading, leading.length, cm.getOption("tabSize")) * charWidth;

				// correction for normal line indentation
				if (_settings.normal_line_indenting && /^\s{1,}[\w\d\"“”'‘’»«].+/.test(line.text)) off = 0;

				elt.style.textIndent = "-" + off + "px";
				elt.style.paddingLeft = (basePadding + off) + "px";
			}
		});
	});



	/*
	########    ###     ########   ########
	##         ## ##    ##     ##  ##
	##        ##   ##   ##     ##  ##
	######   ##     ##  ##     ##  ######
	##       #########  ##     ##  ##
	##       ##     ##  ##     ##  ##
	##       ##     ##  ########   ########
	*/


	CodeMirror.commands.toggleFadeInactiveParagraphs = function (cm) {
		if (localStorage["VIEW_FADE_INACTIVE_PARAGRAPHS"] && localStorage["VIEW_FADE_INACTIVE_PARAGRAPHS"] == "1") {
			localStorage["VIEW_FADE_INACTIVE_PARAGRAPHS"] = "0";
			cm.setOption("fadeInactiveParagraps", false);
			// wm.showNotice("Fade inactive paragraphs is disabled");
		} else {
			localStorage["VIEW_FADE_INACTIVE_PARAGRAPHS"] = "1";
			cm.setOption("fadeInactiveParagraps", true);
			// wm.showNotice("Fade inactive paragraphs is enabled");
		}
	};

	CodeMirror.defineOption("fadeInactiveParagraps", false, function (cm, val, prev) {

		if (prev && prev != CodeMirror.Init) {
			$(".CodeMirror-line").css("opacity", "1");
			cm.off("cursorActivity", fadingRefreshOnCursorMove);
			cm.off("renderLine", renderLineForFade);
		}
		if (val) {
			cm.on("cursorActivity", fadingRefreshOnCursorMove);
			cm.on("renderLine", renderLineForFade);

			$(".CodeMirror-line").css("opacity", _settings.fade_inactive_paragraps_amount);
			cm.refresh();
		}
	});

	var lastActiveLineRendered = -1;

	function renderLineForFade(cm, line, elt) {

		if (line.lineNo() == cm.getCursor("anchor").line) {
			elt.style.opacity = 1;
			lastActiveLineRendered = cm.getCursor().line;
		} else {
			elt.style.opacity = _settings.fade_inactive_paragraps_amount;
		}
	}

	function fadingRefreshOnCursorMove(cm) {
		if (lastActiveLineRendered != cm.getCursor().line) {
			setTimeout(function () {
				cm.refresh()
			}, 10); // instead of that just render this line
		}
	}



	/*
	 ######    ########   ##    ##   ########   ##     ##  ####   ######    ##     ##
	##    ##   ##         ###   ##      ##      ##     ##   ##   ##    ##   ##     ##
	##         ##         ####  ##      ##      ##     ##   ##   ##         ##     ##
	 ######    ######     ## ## ##      ##      #########   ##   ##   ####  #########
	      ##   ##         ##  ####      ##      ##     ##   ##   ##    ##   ##     ##
	##    ##   ##         ##   ###      ##      ##     ##   ##   ##    ##   ##     ##
	 ######    ########   ##    ##      ##      ##     ##  ####   ######    ##     ##
	*/


	CodeMirror.commands.toggleStyleCurrentSentence = function (cm) {
		if (localStorage["VIEW_STYLE_CURRENT_SENTENCE"] && localStorage["VIEW_STYLE_CURRENT_SENTENCE"] == "1") {
			localStorage["VIEW_STYLE_CURRENT_SENTENCE"] = "0";
			cm.setOption("styleCurrentSentence", false);
		} else {
			localStorage["VIEW_STYLE_CURRENT_SENTENCE"] = "1";
			cm.setOption("styleCurrentSentence", true);
		}
	};


	var currentMarker = null;
	// var sentenceDelimitersRegex = new RegExp("[\.\?\!\n]");
	// var sentenceQuotesRegex = new RegExp("[~=_\*\"“”'‘’»«]");

	CodeMirror.defineOption("styleCurrentSentence", false, function (cm, val, prev) {

		if (prev && prev != CodeMirror.Init) {
			cm.off("cursorActivity", higlightCurrentSentence);
			if (currentMarker) currentMarker.clear();
		}
		if (val) {
			cm.on("cursorActivity", higlightCurrentSentence);
			higlightCurrentSentence(cm);
		}
	});


	function higlightCurrentSentence(cm) {
		var cursor = cm.getCursor();
		var line = cm.getLine(cursor.line);
		var sentence = cm.findSentenceAt(line, cursor);

		if (currentMarker) currentMarker.clear();

		// prevent marking sentence if those conditions
		if (cm.somethingSelected()) return; // if selection


		var sentenceMarks = cm.findMarksAt(cursor);
		if (sentenceMarks.some(function (val, index, arr) {
				return val.type == "collapsed_region"
			})) return;


		currentMarker = cm.markText({
			line: cursor.line,
			ch: sentence.start
		}, {
			line: cursor.line,
			ch: sentence.end
		}, {
			className: "CodeMirror-activesentence-background",
			handleMouseEvents: true,
			inclusiveRight: false,
			inclusiveLeft: false
		});

	}

	CodeMirror.defineExtension("findSentenceAt", function (line, pos) {

		var len = line.length;
		var start = 0
		var end = len;

		var regexp = /([^\d][\.\?\!…][\"“”'‘’»«]{0,1}[~=_*/]{0,2} {1,})/g;

		var previousHit = 0;
		var hitBeforeEnd = false;
		var counter = 0;
		var exceptions = " " + _settings.sentence_breakup_dot_exceptions.toLowerCase() + " "; // normalize to lower case
		var beforeSpace;

		while (result = regexp.exec(line)) {

			// test for exceptions
			beforeSpace = line.substring(0, result.index + result.length).trim();
			beforeSpace = beforeSpace.slice(beforeSpace.lastIndexOf(" "));
			//console.log(beforeSpace);

			// check for exceptions
			if (exceptions.indexOf(beforeSpace.toLowerCase()) !== -1) {
				continue;
			}



			var sEnd = result.index + result[0].length;
			counter++;
			if (sEnd >= pos.ch) {
				start = previousHit;
				end = sEnd;
				hitBeforeEnd = true;
				break;
			}
			previousHit = sEnd;

		}

		// correcting for last sentence
		if (!hitBeforeEnd) {
			start = sEnd;
			end = len;
		}

		// no sentences in paragraph, mark whole paragraph
		if (counter == 0) {
			start = 0;
			end = len;
		}

		return {
			start: start,
			end: end,
			text: "" //line.substring(start, end)
		};
	});


	/*

	 ######  ######## ##    ## ########        ########  ########  ##    ##
	##    ## ##       ###   ##    ##           ##     ## ##     ## ##   ##
	##       ##       ####  ##    ##           ##     ## ##     ## ##  ##
	 ######  ######   ## ## ##    ##           ########  ########  #####
	      ## ##       ##  ####    ##           ##     ## ##   ##   ##  ##
	##    ## ##       ##   ###    ##    ###    ##     ## ##    ##  ##   ##
	 ######  ######## ##    ##    ##    ###    ########  ##     ## ##    ##

	*/

	CodeMirror.commands.toggleSentenceBreakup = function (cm) {

		if (localStorage["VIEW_SHOW_SENTENCE_BREAKUP"] && localStorage["VIEW_SHOW_SENTENCE_BREAKUP"] == "1") {
			localStorage["VIEW_SHOW_SENTENCE_BREAKUP"] = "0";
			cm.setOption("showSentenceBreakup", false);
		} else {
			localStorage["VIEW_SHOW_SENTENCE_BREAKUP"] = "1";
			cm.setOption("showSentenceBreakup", true);
		}
	};



	CodeMirror.defineOption("showSentenceBreakup", false, function (cm, val, prev) {

		if (prev == CodeMirror.Init) prev = false;

		if (prev && !val) {
			cm.removeOverlay("wm-sentence-breakup");

		} else if (!prev && val) {
			var beforeSpace;
			var exceptions = " " + _settings.sentence_breakup_dot_exceptions.toLowerCase() + " "; // normalize to lower case
			var endOfSentenceRegEx = /[^\d][\.\?\!…]["“”'‘’»«]{0,1}[~=_*/]{0,2} /;

			cm.addOverlay({
				name: "wm-sentence-breakup",
				token: function (stream, state) {
					//beforeSpace = stream.string.substring(stream.pos - 5, stream.pos).replace(/^.{2,} /, "") + " ";

					beforeSpace = stream.string.substring(0, stream.pos);
					beforeSpace = beforeSpace.slice(beforeSpace.lastIndexOf(" ")) + " ";

					if (stream.match(/^ /) && endOfSentenceRegEx.test(beforeSpace)) {

						// check for exceptions
						if (exceptions.indexOf(beforeSpace.toLowerCase()) !== -1) {
							stream.next();
							return null;
						}

						return "wm-sentence-breakup";
					} else {
						stream.next();
						return null;
					}
				}
			});
		}
	});


	// bionic (anchored) reading

	CodeMirror.commands.toggleAnchoredReading = function (cm) {

		if (localStorage["VIEW_SHOW_ANCHORED_READING"] && localStorage["VIEW_SHOW_ANCHORED_READING"] == "1") {
			localStorage["VIEW_SHOW_ANCHORED_READING"] = "0";
			cm.setOption("showAnchoredReading", false);
		} else {
			localStorage["VIEW_SHOW_ANCHORED_READING"] = "1";
			cm.setOption("showAnchoredReading", true);
		}
	};


	CodeMirror.defineOption("showAnchoredReading", false, function (cm, val, prev) {
        if (prev == CodeMirror.Init) prev = false;
        if (prev && !val) {
            cm.removeOverlay("wm-anchored-reading");
        } else if (!prev && val) {
            cm.addOverlay({
                name: "wm-anchored-reading",
                token: function (stream, state) {
                    var rx_word = "\t!\"#$%&()*+,-./:“”„‘’»«;<=>?¿?@[\\]^_—–`{|}~ ";
                    // var word = "";
                    var wordLength = 0;
                    var minLength = _settings.anchored_reading_min_word_length;
                    var ch = stream.peek();
                    var middleOfTheWordSkip = (!stream.sol() && stream.pos >= _anchoredReading.lastPos && stream.pos < _anchoredReading.lastPos + _anchoredReading.lastBackup);
                    if (rx_word.includes(ch) || middleOfTheWordSkip) {
                        wordLength = 0;
                        stream.next();
                        return null;
                    }
                    while ((ch = stream.peek()) != null && !rx_word.includes(ch)) {
                        wordLength++;
                        stream.next();
                    }
                    if (wordLength >= minLength && _anchoredReading.skippedWords >= _settings.anchored_reading_skip_words) {
                        var back = wordLength - Math.round(wordLength * _settings.anchored_reading_anchor_length);
                        stream.backUp(back);
                        _anchoredReading.lastPos = stream.pos; // set global var for state
                        _anchoredReading.lastBackup = back; // set global var for state
                        _anchoredReading.skippedWords = 0; // set global var for state

                        // console.log(stream.pos);
                        return "wm-anchored-reading";
                    } else {
						_anchoredReading.skippedWords++;
                        stream.next();
                        return null;
                    }
                }
            });
        }
    });






	/*
	##      ## ##     ##  ####  ########   ########    ######    ########
	##  ##  ## ##     ##   ##      ##      ##         ##    ##   ##     ##
	##  ##  ## ##     ##   ##      ##      ##         ##         ##     ##
	##  ##  ## #########   ##      ##      ######      ######    ########
	##  ##  ## ##     ##   ##      ##      ##               ##   ##
	##  ##  ## ##     ##   ##      ##      ##         ##    ##   ##
	 ###  ###  ##     ##  ####     ##      ########    ######    ##
	*/


	// will show white spaces
	CodeMirror.commands.toggleWhiteSpaces = function (cm) {
		cm.setOption("showWhitespaces", !cm.getOption("showWhitespaces"));
	};


	// CodeMirror.commands.toggleWhiteSpaces = function (cm) {

	// 	if (localStorage["VIEW_SHOW_WHITESPACES"] && localStorage["VIEW_SHOW_WHITESPACES"] == "1") {
	// 		localStorage["VIEW_SHOW_WHITESPACES"] = "0";
	// 		cm.setOption("showWhitespaces", false);
	// 	} else {
	// 		localStorage["VIEW_SHOW_WHITESPACES"] = "1";
	// 		cm.setOption("showWhitespaces", true);
	// 	}
	// };

	CodeMirror.defineOption("showWhitespaces", false, function (cm, val, prev) {

		if (prev == CodeMirror.Init) prev = false;

		if (prev && !val) {
			cm.removeOverlay("wm-rspace");
			// remove tabs and new lines
			$("head").append("<style>" + '.' + _currentTheme + " pre.CodeMirror-line > span:after{ display:none }</style>");
			$("head").append("<style>" + '.' + _currentTheme + " .cm-tab:before{ display:none }</style>");

		} else if (!prev && val) {
			// add tabs and new lines
			$("head").append('<style>' + '.' + _currentTheme + ' pre.CodeMirror-line > span:after{ display:inline-block }</style>');
			$("head").append('<style>' + '.' + _currentTheme + ' .cm-tab:before{ display:inline-block}</style>');

			cm.addOverlay({
				name: "wm-rspace",
				token: function (stream, state) {
					if (stream.match(/^ {2,}/) || stream.match(/^ {1,}$/) || (stream.sol() && stream.match(/^ {1,}/))) {
						return "wm-rspace";
					} else {
						stream.next();
						return null;
					}
				}
			});
		}
	});


	/*

	  ######  ##       ########    ###    ##    ##
	 ##    ## ##       ##         ## ##   ###   ##
	 ##       ##       ##        ##   ##  ####  ##
	 ##       ##       ######   ##     ## ## ## ##
	 ##       ##       ##       ######### ##  ####
	 ##    ## ##       ##       ##     ## ##   ###
	  ######  ######## ######## ##     ## ##    ##

	*/
	CodeMirror.commands.cleanText = function (cm) {
		cm.cleanAll();
	};

	CodeMirror.defineExtension("cleanAll", function () {

		var cursor = this.getCursor();
		var isSelection = this.somethingSelected();
		var text = this.getValue();
		var start = {line:0, ch:0};
		var end = {line:this.lastLine(), ch: this.getLine(this.lastLine()).length};

		if (isSelection) {
			var sel = this.listSelections();
			text = this.getSelection();
			start = sel[0].head;
			end = sel[0].anchor;
		}

		var text = (isSelection) ? this.getSelection() : this.getValue();

		text = this.cleanWhiteSpaces(text);
		text = this.cleanNewLines(text, isSelection);
		text = (_settings.clean_text_straighten_quotes == true) ? this.cleanCharacters(text) : text;

		this.replaceRange(text, start, end);

		if (!isSelection) {
			jumps.jumpTo(cursor.line); // better save scroll pos and scroll back
		}

		wm.showNotice("<strong>Text cleaned!</strong><br><br>All redundant spaces, tabs and newlines have been removed. " + ((_settings.clean_text_straighten_quotes == true) ?  "Quotes have been straightened." : ""));

	});

	// Clean newlines
	CodeMirror.defineExtension("cleanWhiteSpaces", function (text) {

		text = text
		.replace(/^[ \t]+$/gm, "") // remove spaces or tabs from blank lines
		.replace(/[ \t]+$/gm, "") // remove trailing spaces or tabs at the end of not empty lines
		.replace(/(\S) {2,}/gm, "$1 "); // replace 2 or more spaces inside text with single space but not if at the beginning of not empty line

		return text;
	});


	// Clean newlines
	CodeMirror.defineExtension("cleanNewLines", function (text, isSelection) {

		text = text
		.replace(/\r\n|\n\r|\n|\r/g, "\n") // normalize newlines
		.replace(/\n{3,}/g, "\n\n"); // replace 3 or more newlines with 2

		if (!isSelection) {
			text = text.replace(/\n+$/, ""); // remove trailing new lines if whole text
		}

		return text;
	});


	// Clean newlines
	CodeMirror.defineExtension("cleanCharacters", function (text) {

		text = text
		.replace(/‘|’|′/g, "'").replace(/“|”|«|»|″|„/g, '"').replace(/‴/g, "'''") // straighten quote marks
		.replace(/…/g, "...").replace(/\.{4,}/g, "..."); // replace / shorten ellipsis

		return text;
	});



	/*
	##     ##     ###     ########    ######    ##     ##  ########    ######
	###   ###    ## ##       ##      ##    ##   ##     ##  ##         ##    ##
	#### ####   ##   ##      ##      ##         ##     ##  ##         ##
	## ### ##  ##     ##     ##      ##         #########  ######      ######
	##     ##  #########     ##      ##         ##     ##  ##               ##
	##     ##  ##     ##     ##      ##    ##   ##     ##  ##         ##    ##
	##     ##  ##     ##     ##       ######    ##     ##  ########    ######
	*/
	// Will show matches on scrollbar and add overlay (searching) for that query
	// show hatches / remove matches
	var wmMatchesOnScrollbar = null;

	CodeMirror.defineExtension("showMatches", function (str) {
		var query = new RegExp(escapeRegExp(str), "i");

		var overlayO = {
			name: "searching",
			token: function (stream) {

				if (stream.match(query)) {
					return "searching";
				} else {
					stream.next();
					return null;
				}

			}
		};
		this.addOverlay(overlayO);

		wmMatchesOnScrollbar = this.showMatchesOnScrollbar(str, true, {});
	});


	CodeMirror.defineExtension("removeMatches", function () {
		if (wmMatchesOnScrollbar) {
			wmMatchesOnScrollbar.clear();
			wmMatchesOnScrollbar = null;
		}
		this.removeOverlay("searching");
	});


	CodeMirror.commands.showSelectionMatches = function (cm) {

		if (wmMatchesOnScrollbar) {
			cm.removeMatches();

		} else {

			if (cm.somethingSelected()) {
				var q = cm.getSelection();
			} else {
				var w = cm.findWordAt(cm.getCursor());
				var q = cm.getRange(w.from(), w.to());
			}

			cm.execCommand("clearSearch");
			if (q.length >= 1) {
				cm.showMatches(q);
			}
		}
	};


	/*
	########    #######    #######   ##     ##
		 ##    ##     ##  ##     ##  ###   ###
		##     ##     ##  ##     ##  #### ####
	   ##      ##     ##  ##     ##  ## ### ##
	  ##       ##     ##  ##     ##  ##     ##
	 ##        ##     ##  ##     ##  ##     ##
	########    #######    #######   ##     ##
	*/



	// Defines logic for text zooming
	CodeMirror.defineInitHook(function (cm) {

		function setTextSize(value, isDelta) {
			var ts = parseInt($(".CodeMirror").css('font-size'));
			var nts = (isDelta) ? (ts + value) : value;
			nts = (nts <= 3) ? 3 : nts;

			$(".CodeMirror").css("font-size", nts + "px");
			cm.scrollIntoView(null, cm.getWrapperElement().offsetHeight / 3);
			localStorage['VIEWPORT_CURRENT_FSIZE'] = nts;
			cm.refresh();
		}

		function setTextAreaWidth(value, isDelta) {
			var w = getWidthInPercent($(".editor-placeholder"));
			var nw = (isDelta) ? w + value : value;
			if (nw >= 96) {
				nw = 96;
			} else if (nw <= 20) {
				nw = 20;
			}

			$(".editor-placeholder").css("width", nw + "%");
			cm.scrollIntoView(null, cm.getWrapperElement().offsetHeight / 3);
			localStorage['VIEWPORT_CURRENT_WIDTH'] = nw;
			cm.refresh();
		}

		function zoom(delta) {
			// setTextSize(delta);
			// setTextAreaWidth(delta * 5);
			// cm.scrollIntoView(null, cm.getWrapperElement().offsetHeight / 3);
		}

		function nudge(delta) {

		}

		function resZoom() {

			if (win.isFullscreen) {
				var W = localStorage["VIEWPORT_DEFAULT_WIDTH_FS"] ? localStorage["VIEWPORT_DEFAULT_WIDTH_FS"] : 70;
				var F = localStorage["VIEWPORT_DEFAULT_FSIZE_FS"] ? localStorage["VIEWPORT_DEFAULT_FSIZE_FS"] : 17;

			} else {
				var W = localStorage["VIEWPORT_DEFAULT_WIDTH_WIN"] ? localStorage["VIEWPORT_DEFAULT_WIDTH_WIN"] : 100;
				var F = localStorage["VIEWPORT_DEFAULT_FSIZE_WIN"] ? localStorage["VIEWPORT_DEFAULT_FSIZE_WIN"] : 14;
			}

			setTextAreaWidth(W, false);
			setTextSize(F, false);
		}


		function setResetZoomDefault() {

			var settW = (win.isFullscreen) ? "VIEWPORT_DEFAULT_WIDTH_FS" : "VIEWPORT_DEFAULT_WIDTH_WIN";
			var settF = (win.isFullscreen) ? "VIEWPORT_DEFAULT_FSIZE_FS" : "VIEWPORT_DEFAULT_FSIZE_WIN";

			// save values
			localStorage[settW] = getWidthInPercent($(".editor-placeholder"));
			localStorage[settF] = parseInt($(".CodeMirror").css('font-size'));
		}



		function onWheel(e) {
			var direction = (e.wheelDeltaY > 0) ? 1 : -1;
			if (e.altKey) {
				e.preventDefault();
				setTextAreaWidth(5 * direction, true);
			}
			if (e.ctrlKey || e.metaKey) {
				e.preventDefault();
				setTextSize(1 * direction, true);
			}
		}



		CodeMirror.on(document, "mousewheel", onWheel);
		// CodeMirror.on(document, "DOMMouseScroll", onWheel);

		CodeMirror.commands.increaseFontZoom = function (cm) {
			setTextSize(1, true);
		};

		CodeMirror.commands.decreaseFontZoom = function (cm) {
			setTextSize(-1, true);
		};

		CodeMirror.commands.increaseTextAreaWidth = function (cm) {
			setTextAreaWidth(5, true);
		};

		CodeMirror.commands.decreaseTextAreaWidth = function (cm) {
			setTextAreaWidth(-5, true);
		};

		CodeMirror.commands.initZoom = function (cm) {
			if (localStorage['VIEWPORT_CURRENT_FSIZE']) setTextSize(localStorage['VIEWPORT_CURRENT_FSIZE'], false);
			if (localStorage['VIEWPORT_CURRENT_WIDTH']) setTextAreaWidth(localStorage['VIEWPORT_CURRENT_WIDTH'], false);
		};


		CodeMirror.commands.resetZoom = function (cm) {
			resZoom();
		};

		CodeMirror.commands.setResetZoomDefault = function (cm) {
			setResetZoomDefault();
		};



	});

	/*

	##       #### ##    ## ##    ##  ######
	##        ##  ###   ## ##   ##  ##    ##
	##        ##  ####  ## ##  ##   ##
	##        ##  ## ## ## #####     ######
	##        ##  ##  #### ##  ##         ##
	##        ##  ##   ### ##   ##  ##    ##
	######## #### ##    ## ##    ##  ######

	*/


	// Make links clickable
	CodeMirror.defineInitHook(function (cm) {

		cm.on("mousedown", function (cm, e) {
			// var modKey = (os.platform() === 'darwin') ? e.shiftKey : e.shiftKey;
			var et = $(e.target);
			if ((et.hasClass("cm-url") || et.hasClass("cm-link")) && e.shiftKey == true) {
				e.codemirrorIgnore = true;
				var link = e.srcElement.innerText;
				opener(link);
			}
		});

	});



	// search in google
	CodeMirror.commands.searchWithGoogle = function (cm) {
		var q = "";
		if (cm.somethingSelected()) { // get selection
			q = cm.getSelection();
		} else {
			q = cm.getLine(cm.getCursor().line); // get current line text
		}
		if (q.trim() !== "") {
			opener('https://www.google.com/search?q=' + encodeURI(q.trim()));
		}
	};



	/*
	##     ##  ####  ########   ########    ######     ######    ########
	##     ##   ##   ##     ##  ##         ##    ##   ##    ##   ##     ##
	##     ##   ##   ##     ##  ##         ##         ##         ##     ##
	#########   ##   ##     ##  ######      ######    ##         ########
	##     ##   ##   ##     ##  ##               ##   ##         ##   ##
	##     ##   ##   ##     ##  ##         ##    ##   ##    ##   ##    ##
	##     ##  ####  ########   ########    ######     ######    ##     ##
	*/
	// Defines scroll bar hiding
	CodeMirror.defineInitHook(function (cm) {



		function showScrollBar() {

			if ($(".CodeMirror-overlayscroll-vertical").width() <= 0) {
				$(".CodeMirror-overlayscroll-vertical").css("width", "6px");
			}

		}

		function hideScrollBar() {
			if ($(".CodeMirror-overlayscroll-vertical").width() > 0) {
				$(".CodeMirror-overlayscroll-vertical").animate({
					"width": "-1px"
				}, 150);
			}
		}


		var hideTimer;

		function resetHideTimer() {
			clearInterval(hideTimer);
			hideTimer = setInterval(hideScrollBar, 3000);
		}


		function onScroll(e) {
			resetHideTimer();
			showScrollBar();
		}

		function onChange(e) {
			clearInterval(hideTimer);
			hideScrollBar();

			// set not saved indicator
			if (_isEditorSaved) {
				_isEditorSaved = false;
				storage.uiUpdateDocumentName(_currentDoc.nam);
			}
		}



		CodeMirror.on(cm.getWrapperElement(), "mousewheel", onScroll);
		// CodeMirror.on(cm.getWrapperElement(), "DOMMouseScroll", onScroll);
		cm.on("change", onChange);
		cm.on("scroll", resetHideTimer);

		CodeMirror.on(cm.getWrapperElement(), "keyup", function (e) {
			if ((e.keyCode == 38 || e.keyCode == 40 || e.keyCode == 33 || e.keyCode == 34) &&
				(!e.shiftKey && !e.altKey)) {
				onScroll(e);
			}
		});



		CodeMirror.commands.showScrollBar = function (cm) {
			showScrollBar();
			resetHideTimer()
		};
		// CodeMirror.commands.hideScrollBar = function(cm) { };


	});


	/*
	##     ##  ##    ##   ########      ###      ######     ######
	###   ###  ##   ##       ##        ## ##    ##    ##   ##    ##
	#### ####  ##  ##        ##       ##   ##   ##         ##
	## ### ##  #####         ##      ##     ##  ##   ####   ######
	##     ##  ##  ##        ##      #########  ##    ##         ##
	##     ##  ##   ##       ##      ##     ##  ##    ##   ##    ##
	##     ##  ##    ##      ##      ##     ##   ######     ######
	*/
	CodeMirror.commands.toggleBold = function (cm) {
		var c = _settings.italic_bold_character;

		cm.toggleMarks(c + c, c + c, false, false);
	};
	CodeMirror.commands.toggleItalic = function (cm) {
		var c = _settings.italic_bold_character;
		cm.toggleMarks(c, c, false, false);
	};
	CodeMirror.commands.toggleStrikethrough = function (cm) {
		cm.toggleMarks("~~", "~~", false, false);

	};
	CodeMirror.commands.toggleComment = function (cm) {
		cm.toggleMarks("//", "//", true, true);
	};

	CodeMirror.commands.toggleCode = function (cm) {
		cm.toggleMarks("`", "`", false, false);
	};

	CodeMirror.commands.toggleMarked = function (cm) {
		cm.toggleMarks("==", "==", false, false);
	};

	// headings
	CodeMirror.commands.headingsDown = function (cm) {
		var hCursor = cm.getCursor();
		var hLevel = cm.getHeaderLevel(hCursor);
		var hLine = cm.getLine(hCursor.line);

		if (hLevel == -1) return;

		var newLine = hLine.substr(1).trim();

		var cursorCorr = 1;
		if (hLevel == 1) {
			cursorCorr = 2;
		}
		cm.replaceRange(newLine, {
			line: hCursor.line,
			ch: 0
		}, {
			line: hCursor.line,
			ch: hLine.length
		});
		cm.setCursor({
			line: hCursor.line,
			ch: hCursor.ch - cursorCorr
		});

	};

	CodeMirror.commands.headingsUp = function (cm) {
		var hCursor = cm.getCursor();
		var hLevel = cm.getHeaderLevel(hCursor);
		var hLine = cm.getLine(hCursor.line);

		if (hLevel >= 6) return;

		var newLine = "";
		var cursorCorr = 1;
		if (hLevel == -1) {
			newLine = '# ' + hLine;
			cursorCorr = 2;
		} else {
			newLine = '#' + hLine;
		}
		cm.replaceRange(newLine, {
			line: hCursor.line,
			ch: 0
		}, {
			line: hCursor.line,
			ch: hLine.length
		});
		cm.setCursor({
			line: hCursor.line,
			ch: hCursor.ch + cursorCorr
		});
	};

	CodeMirror.commands.convertToMdTable = function (cm) {
		
		if (cm.somethingSelected()) {
			var q = cm.getSelection();
		} else {
			wm.showNotice("You must first select lines with tab-separated data!");
			return;
		}

		var tp = {
			edgeL: "| ",
			edgeR: " |\n",
			inside: " | ",
			hdiv: "|:--"
		}
		var table = "";
		var lines = q.split("\n");

		var numOfC = (lines[0].match(/\t{1,}/g) || []).length;
		for (var i = 0; i < lines.length; i++) {
			table += tp.edgeL + lines[i].replace(/\t{1,}/g, tp.inside) + tp.edgeR;
			if (i == 0) { // first line, add divider after
				table += tp.hdiv.repeat(numOfC + 1) + "|\n";
			}
		}
		cm.replaceSelection(table);

	};


	CodeMirror.defineExtension("toggleMarks", function (startMark, endMark, startOnly, spaced) {
		var cm = this;

		function escapeRegExp(string) {
			return string.replace(/([.*+?^${}()|\[\]\/\\])/g, "\\$1");
		}

		function toggle(text, rangeToAct, isLine) {
			var regexLeft = new RegExp("^\\s*" + escapeRegExp(startMark) + "\\s*");
			var regexRight = new RegExp("\\s*" + escapeRegExp(endMark) + "\\s*$");
			var newText;
			var space = (spaced && isLine) ? " " : "";

			if (regexLeft.test(text)) {
				newText = text.replace(regexLeft, ""); // remove start
			} else {
				newText = startMark + space + text; // add start
			}
			if (regexRight.test(text)) {
				newText = newText.replace(regexRight, ""); // remove end
			} else {
				newText = newText + ((startOnly && isLine) ? "" : endMark); // add end
			}
			cm.replaceRange(newText, rangeToAct.anchor, rangeToAct.head);

			if (!startOnly) {
				cm.addSelection(rangeToAct.anchor, {
					line: rangeToAct.anchor.line,
					ch: rangeToAct.anchor.ch + newText.length
				});
			}
		}



		cm.operation(function () {
			var selectionsA = cm.listSelections();


			for (var i = selectionsA.length - 1; i >= 0; i--) { // loop all selections and cursors
				var from = selectionsA[i].from();
				var to = selectionsA[i].to();
				var rangeToAct;
				var isLine;
				if (from.line == to.line) { // spans single line
					isLine = true;

					if (from.ch == to.ch) { // cursor only

						// ceck if cursor and inside existing block
						// find for marks before and after
						var testCur = cm.getCursor();
						var testLine = cm.getLine(testCur.line);


						var startPos = testLine.substring(0, testCur.ch).lastIndexOf(startMark);
						var endPos = testLine.indexOf(endMark, testCur.ch);

						if (endPos > testCur.ch && startPos != -1 && startPos < testCur.ch) {
							rangeToAct = { // get existing marks range from cursor
								anchor: {
									line: from.line,
									ch: startPos
								},
								head: {
									line: from.line,
									ch: endPos + endMark.length
								}
							};
						} else {
							rangeToAct = { // get line range from cursor
								anchor: {
									line: from.line,
									ch: 0
								},
								head: {
									line: from.line,
									ch: cm.getLine(from.line).length
								}
							};
						}

					} else { // selection
						isLine = false;
						rangeToAct = {
							anchor: from,
							head: to
						};
					}
					// Now toggle
					var text = cm.getRange(rangeToAct.anchor, rangeToAct.head);
					toggle(text, rangeToAct, isLine); // call toggle functions


				} else { // spans multiple lines
					for (var j = to.line; j >= from.line; j--) { // loop lines in selection, act on whole lines
						var lineText = cm.getLine(j);
						if (lineText <= 0) continue; // skip if line is empty
						rangeToAct = { // get line range from cursor
							anchor: {
								line: j,
								ch: 0
							},
							head: {
								line: j,
								ch: lineText.length
							}
						};
						// Now toggle
						var text = cm.getRange(rangeToAct.anchor, rangeToAct.head);
						toggle(text, rangeToAct, true); // call toggle functions
					}
					cm.execCommand("goLineStart");
				}
			} // end loop
			cm.refresh();
		});
	});



	// Temporary redefine word bounds to include * an ~ ...
	// CodeMirror.registerHelper("wordChars", "markdown", /[\w$\*~\/\`]/);

	// Reset to standard word bounds
	// CodeMirror.registerHelper("wordChars", "markdown", /[\w$]/);


	/*
	########   ########   ########    #######
	##     ##  ##         ##     ##  ##     ##
	##     ##  ##         ##     ##  ##     ##
	########   ######     ########   ##     ##
	##   ##    ##         ##         ##     ##
	##    ##   ##         ##         ##     ##
	##     ##  ########   ##          #######
	*/


	CodeMirror.commands.moveToRepository = function (cm) {
		cm.sendToRepository(cm.getSelectedOrLines(true));
		if (!universalPane.isPaneVisible("wm_repo")) wm.showNotice("Text moved to repository");
	};


	CodeMirror.commands.copyToRepository = function (cm) {
		cm.sendToRepository(cm.getSelectedOrLines(false));
		if (!universalPane.isPaneVisible("wm_repo")) wm.showNotice("Text copied to repository");
	};


	CodeMirror.defineExtension("sendToRepository", function (text) {
		var cm = this;
		// get selections and cursors / iterate over
		//var selectionsA = cm.listSelections();
		repository.insertNewRepoItem(text, {});
	});




	/*
	 ######    ########   ########   ##         ##
	##    ##   ##     ##  ##         ##         ##
	##         ##     ##  ##         ##         ##
	 ######    ########   ######     ##         ##
	      ##   ##         ##         ##         ##
	##    ##   ##         ##         ##         ##
	 ######    ##         ########   ########   ########
	*/



	// Define some logic for spelling
	CodeMirror.defineInitHook(function (cm) {



		CodeMirror.commands.toggleSpelling = function (cm) {
			var dic = (localStorage["LANGUAGE_SPELLING_DICTIONARY"]) ? localStorage["LANGUAGE_SPELLING_DICTIONARY"] : "en_US";

			if (cm.getOption("mode") == "spell-checker") {
				cm.setOption("mode", "gfm");
				console.log("Disabling spelling for: " + dic);
			} else {
				cm.setDictionary(dic);
				cm.setOption("backdrop", "gfm");
				cm.setOption("mode", "spell-checker");
				console.log("Enabling spelling for: " + dic);
			}
		};



		CodeMirror.defineExtension("setDictionary", function (dicName) {
			var cm = this;
			dicName = dicName.replace(/\..*$/, "");
			localStorage["LANGUAGE_SPELLING_DICTIONARY"] = dicName;

			var currentDic = "";
			if (typo) {
				// currentDic = mySpell;
				// console.log("Current dic: ", currentDic)
				// if (currentDic == "get name from instance") {
				// 	console.log("Already active: " + dicName);
				// 	return;
				// } else


				if (loadedDictionaries[dicName]) {
					console.log("Switching to existing lang: " + dicName);
					typo = loadedDictionaries[dicName];
					cm.setOption("backdrop", "gfm");
					cm.setOption("mode", "spell-checker");
					return;
				}
			}
			// if code gets past this ...
			console.log("Loading dic from scratch: " + dicName);

			cm.loadDictionary(dicName);
			cm.setOption("backdrop", "gfm");
			// cm.setOption("mode", "spell-checker");
			setTimeout(function () {
				cm.setOption("mode", "spell-checker")
			}, 1500);

		});


		CodeMirror.defineExtension("loadCustomDictionary", function () {

			var dbPath = (_settings.data_directory != "") ? _settings.data_directory : app.dataPath;

			fs.readFile(dbPath + "/" + "writemonkey3_custom_dictionary", "utf-8", function (err, data) {
				if (err) return;

				var words = data.split("\n");
				for (var i = 0; i < words.length; i++) {
					ignoreList.push(words[i].trim());
				}
				console.log("Custom dictionary loaded.")
			});
		});

		CodeMirror.defineExtension("loadDictionary", function (dicName) {
			// read aff file
			fs.readFile("./lang/" + dicName + ".aff", "utf-8", function (err, data) {
				if (err) return;
				aff_data = data;
				num_loaded++;
				if (num_loaded == 2) {
					num_loaded = 0;

					typo = new Typo(dicName, aff_data, dic_data, {
						platform: 'any'
					});
					loadedDictionaries[dicName] = typo;

					// mySpell = nspell(aff_data, dic_data);
					// loadedDictionaries[dicName] = mySpell;
					// console.log(mySpell);
				}
			});


			// read dic file
			fs.readFile("./lang/" + dicName + ".dic", "utf-8", function (err, data) {
				if (err) return;
				dic_data = data;
				num_loaded++;
				if (num_loaded == 2) {
					num_loaded = 0;

					typo = new Typo(dicName, aff_data, dic_data, {
						platform: 'any'
					});
					loadedDictionaries[dicName] = typo;

					// mySpell = nspell(aff_data, dic_data);
					// loadedDictionaries[dicName] = mySpell;
					// console.log(mySpell);

				}
			});

		});


		CodeMirror.defineExtension("getSpellingDictionaries", function (dicName) {
			var filesA = fs.readdirSync("lang/");

			// construct palette items
			var objectsA = [];

			for (var i = 0; i < filesA.length; i++) {
				if (filesA[i].endsWith(".dic")) {
					var item = {
						string: "Set dictionary: " + filesA[i].replace(/\..*$/, ""),
						shortcut: "",
						action: '_editor.setDictionary("' + filesA[i] + '")',
						type: "eval",
						sublist: [],
						persist: false
					};
					objectsA.push(item);
				}
			}
			return objectsA;

		});



		CodeMirror.on(cm.getWrapperElement(), "contextmenu", function (e) {

			function addWordToDictionary(word) {
				var dbPath = (_settings.data_directory != "") ? _settings.data_directory : app.dataPath;
				fs.appendFile(dbPath + '/' + 'writemonkey3_custom_dictionary', word + '\n', {
					encoding: 'utf8'
				}, function (err) {
					addToIgnoreList(word);
				});
			}


			function addToIgnoreList(word) {
				if (ignoreList.indexOf(word) == -1) {
					ignoreList.push(word);
					// re-run mode
					cm.setOption("backdrop", "gfm");
					cm.setOption("mode", "spell-checker");
				}
			}


			function getSuggestions(word) {
				return typo.suggest(word);
				// return mySpell.suggest(word);
			}

			function getSuggestionsContextMenu(word) {
				var suggestionsA = getSuggestions(word);
				//suggestContextMenu.remove();
				suggestContextMenu = null;
				suggestContextMenu = new gui.Menu({
					'type': 'menubar'
				});

				if (suggestionsA.length < 1) {
					var noSuggestions = new gui.MenuItem({
						enabled: false,
						label: "No suggestions found"
					});
				}

				var addToDictionary = new gui.MenuItem({
					enabled: true,
					label: "Learn spelling",
					click: function () {
						addWordToDictionary(cm.getSelection().trim());
					}
				});

				var ignoreWord = new gui.MenuItem({
					label: "Ignore spelling",
					click: function () {
						addToIgnoreList(cm.getSelection());
					}
				});


				for (var i = 0; i < suggestionsA.length; i++) {

					suggestContextMenu.append(new gui.MenuItem({
						label: suggestionsA[i],
						click: function () {
							cm.replaceSelection(this.label, "around");
						}
					}));
				}

				noSuggestions && suggestContextMenu.append(noSuggestions);
				suggestContextMenu.append(new gui.MenuItem({
					type: 'separator'
				}));

				suggestContextMenu.append(ignoreWord);
				suggestContextMenu.append(addToDictionary);


				return suggestContextMenu;
			}


			var token = cm.getTokenAt(cm.getCursor());

			var line = cm.getCursor().line;
			var start = {
				line: line,
				ch: token.start
			};
			var end = {
				line: line,
				ch: token.end
			};
			if (token.type && token.type.indexOf("spell-error") > -1) {
				e.preventDefault();

				cm.setSelection(start, end);
				var menu = getSuggestionsContextMenu(token.string);
				menu.popup(e.clientX, e.clientY);
			}
		});


		/*
		 ######## ########  ######  ########
			##    ##       ##    ##    ##
			##    ##       ##          ##
			##    ######    ######     ##
			##    ##             ##    ##
			##    ##       ##    ##    ##
			##    ########  ######     ##
		*/

		CodeMirror.defineExtension("testFunction", function () {

		});


	});
});