var universalPane = (function() {


	// also add support for bottom pane

	var registeredPanes = {};
	var currentPaneId = "";

	// for resizing
	var isResizing = false;
    var lastDownX = 0;



	var paneMenu;


	function _init() {
		testPanes();

		initPane();
		setPanesMenu();
	}


	function testPanes() {}


	function initPane() {
		var p = localStorage["UNIVERSALPANE_RIGHT_ID"] ? localStorage["UNIVERSALPANE_RIGHT_ID"] : "wm_repo";
		switchPane(p);
	}



	function registerPane(id, name, options) {


		// process options
		var defaultOptions = {
			location: "right",
			width: 350,
			backgroundColor: "transparent",
			textColor: "inherit",
			overflow: "hidden"
		};

		if (typeof options == 'object') {
			options = $.extend(defaultOptions, options);
		} else {
			options = defaultOptions;
		}

		// check if id exists

		registeredPanes[id] = {
			id: id,
			name: name,
			width: options.width,
			backgroundColor: options.backgroundColor,
			textColor: options.textColor,
			location: options.location,
			handle: createNewPane(id, name, options)
		}

		return id;
	}



	function createNewPane(id, name, options) {
		console.log("creating pane");

		// pass options


		// create and store dom element
		var newPane = $('<div/>', {
			id: id,
			type: "right", // bottom
			class: "universal-pane-container",
			style: 'display: block; background-color: ' + options.backgroundColor +  '; color: ' + options.textColor + '; overflow: ' + options.overflow + ';',
			html: ""
		});


		// append to existing dom
		switch (options.location) {
			case "left":
				newPane.appendTo('#module_main');
			case "bottom":
				newPane.appendTo('#util_pane_bottom');
			default:
				newPane.appendTo('#module_main');
		}

		return newPane;
	}


	function setPanesMenu() {
		paneMenu = new gui.Menu({ // rebuild menu
			'type': 'menubar'
		});

		for (prop in registeredPanes) {

			var id = registeredPanes[prop].id;
			var name = registeredPanes[prop].name;



			var menuItem = new gui.MenuItem({
				label: name,
				checked: (id == currentPaneId),
				type: "checkbox",
				tooltip: id
			});
			menuItem.click = function() {
				switchPane(this.tooltip);
			}

			paneMenu.append(menuItem);
		}
	}



	// function(id) {
	// 					switchPane(id);
	// 					console.log(id);
	// 				}


	function getPaneDomHandle(id) {
		return $("#" + id);
	}



	function isPaneVisible(id) {
		return (id == currentPaneId && $("#module_pane").is(':visible') && $("#module_pane").width() > 1) ? true : false;
	}
	

	function switchPaneAndShow(id) {
		switchPane(id);
		$("#module_pane").show();
		localStorage.VIEWPORT_MODULE_PANE_VISIBLE = 1;
	}


	function switchPane(id) {

		// hide all panes
		$(".universal-pane-container").css({
			"display": "none"
		});

		
		// check if id is registered
		if (registeredPanes.hasOwnProperty(id)) {
			// show panel
			
			// get width / default or 
			var pane_width = localStorage['UNIVERSALPANE_WIDTH_' + id.toUpperCase()];
			pane_width = pane_width ? pane_width : registeredPanes[id].width;

			$("#" + id).show();

			$("#module_pane").css({
				"flex": "0 0 " + pane_width + "px"
			});

			$("#module_header, #module_footer").css({
				"background-color": $("#" + id).css("background-color"),
			});
			

			// save setting
			localStorage.setItem('UNIVERSALPANE_RIGHT_ID', id);

			// fire event
			var switchPaneE = new CustomEvent("switchPaneE", {
				detail: {
					oldId: currentPaneId,
					newId: id
				}
			});
			document.body.dispatchEvent(switchPaneE);

			currentPaneId = id;
			_editor.refresh();
			

			// var event = jQuery.Event( "logged" );
			// event.user = "foo";
			// event.pass = "bar";
			// $( "body" ).trigger( event );

		} else {
			// show default (existing pane)
		}
	}

	

	function togglePanes() {
		if (localStorage.VIEWPORT_MODULE_PANE_VISIBLE == 0) {
			$("#module_pane").show();
			localStorage.VIEWPORT_MODULE_PANE_VISIBLE = 1;
			_editor.refresh();
			return;
		}

		var pA = Object.keys(registeredPanes);
		var curr =  localStorage["UNIVERSALPANE_RIGHT_ID"] ?  pA.indexOf(localStorage["UNIVERSALPANE_RIGHT_ID"]) : 0;
		var next = (curr + 1 >= pA.length) ? 0 : curr + 1;

		if (curr + 1 >= pA.length) {
			$("#module_pane").hide();
			localStorage.VIEWPORT_MODULE_PANE_VISIBLE = 0;
		}

		switchPane(pA[next]);
	}



	/*
	########   ##     ##  ########   ##    ##   ########    ######    
	##         ##     ##  ##         ###   ##      ##      ##    ##   
	##         ##     ##  ##         ####  ##      ##      ##         
	######     ##     ##  ######     ## ## ##      ##       ######    
	##          ##   ##   ##         ##  ####      ##            ##   
	##           ## ##    ##         ##   ###      ##      ##    ##   
	########      ###     ########   ##    ##      ##       ######    
	*/

	$("#module_selector").click(function(e) {
		setPanesMenu();
		paneMenu.popup(e.originalEvent.x, e.originalEvent.y + 10);
	});

	
	// mouse resizing
	var mp = $("#module_pane");

	mp.on('mousedown', function(e) {
		if (e.pageX <= (mp.offset().left + 5)) {
			isResizing = true;
			lastDownX = e.clientX;
		}
	});

	mp.on('mousemove', function(e) {
		if (e.pageX <= (mp.offset().left + 5)) {
			mp.css('cursor', 'col-resize');
		} else {
			mp.css('cursor', 'default');
		}
	});

	$(document).on('mousemove', function(e) {
		if (!isResizing)
			return;

		var offsetRight = mp.width() - (e.clientX - mp.offset().left);
		mp.css('flex', '0 0 ' + offsetRight + 'px');
	});



	$(document).on('mouseup', function(e) {
		if (isResizing) {
			localStorage['UNIVERSALPANE_WIDTH_' + currentPaneId.toUpperCase()] = mp.width() - (e.clientX - mp.offset().left);
			_editor.refresh();
		}
		isResizing = false;
	});




	/*
	   ###     ########   ####  
	  ## ##    ##     ##   ##   
	 ##   ##   ##     ##   ##   
	##     ##  ########    ##   
	#########  ##          ##   
	##     ##  ##          ##   
	##     ##  ##         ####  
	*/
	return {
		_init: _init,
		registerPane: registerPane,
		isPaneVisible: isPaneVisible,
		getPaneDomHandle: getPaneDomHandle,
		togglePanes: togglePanes,
		switchPane: switchPane,
		switchPaneAndShow: switchPaneAndShow
	};

})();