module.exports = function (api, pluginInfo) {

	var options = pluginInfo.options;
	var $ = api.window.$;
	var m = window;
	var https = require("https");

	var isSelection;
	var identity;

	function _init() {
		setPaletteCommands();
		setShortcut();
	}

	function setPaletteCommands() {
		m.commandPalette.registerPaletteItem({
			id: 'ai_helper',
			string: 'AI Helper: Ask Artificial Inteligence (OpenAI) for Help',
			shortcut: "Ctrl+Alt+I",
			action: '_plugins["ai-helper"].getAnswer()',
			type: 'eval',
			sublist: [],
			persist: false,
			icon: "&#xf0d0;",
			widget: {
				type: "eval",
				value: '"Prompt is a selection or a current line"'
			}
		});
	}

	function setShortcut() {
		m._editor.on("keydown", function (cm, e) {
			if (e.keyCode == 73 && e.altKey && (e.ctrlKey || e.metaKey)) {
				getAnswer();
			}
		});
	}

	function getAnswer() {

		// exit if no api key
		if (options.openAiApiKey == "") {
			m.wm.showNotice("To use this plugin, you must first get yourself a free account at https://chat.openai.com. Create an API key and paste it into plugins/AI_helper/package.json (openAiApiKey field).", {
				autoClose: false
			});
			return;
		}

		m.wm.showNotice("<strong>AI is preparing an answer. </strong>\n\nWait for it ...", {});
		var p = getPrompt();
		setIdentityFromPrompt(p);
		callApiNodeHttps(p);
	}

	function getPrompt() {
		var t = "";
		if (m._editor.somethingSelected()) { // get selection
			t = m._editor.getSelection();
			isSelection = true;
		} else {
			t = m._editor.getLine(m._editor.getCursor().line); // get current line text
			isSelection = false;
		}
		return t.replace(/^\/{2}\s*/, "");
	}


	function setIdentityFromPrompt(prompt) {
		var idMatch = prompt.trim().match(/^\/(\w+)\s+/);
		if (idMatch && typeof options.AiIdentities[idMatch[1].toLowerCase()] !== 'undefined') { // check if there is identity with that name
			identity = options.AiIdentities[idMatch[1].toLowerCase()];
			identity.name = idMatch[1].toLowerCase();

		} else { // set default
			identity = options.AiIdentities["default"];
			// identity.settings.conversation_id = 50;
			identity.name = "default";
		}

		// console.log(identity);
	}


	function callApiNodeHttps(prompt) {
		var data = "";
		var opt = {
			hostname: options.openAiEndPointUrl,
			path: options.openAiEndPointPath,
			method: 'POST',
			headers: {
				'Content-Type': 'application/json',
				'Authorization': `Bearer ${options.openAiApiKey}`
			}
		};
		var req = https.request(opt, function (res) {
			res.setEncoding('utf8');
			res.on('data', function(chunk) {
				data += chunk;
			});

			res.on('error', function (err) {
				m.wm.showNotice("<strong>AI Helper Plugin Error:</strong>\n\n" + err.toString(), {
					autoClose: false
				});
			});

			res.on('end', function () {
				// console.log(data);
				var response = (m.JSON.parse(data).choices[0].text);
				if (!isSelection) { // cursor at the end of the line
					m._editor.setCursor({
						line: m._editor.getCursor().line,
						ch: m._editor.getLine(m._editor.getCursor().line).length
					});
				}
				m._editor.replaceSelection((isSelection ? prompt : "") + "\n\n");
				m._editor.replaceSelection(response.trim(), "around");
				m.wm.showNotice("<strong>AI finished thinking and gave a response.</strong>\n\nIdentity used: " + identity.name.toUpperCase() + "\ntemp: " + identity.settings.temperature.toString() + "\n\nCost was " + m.JSON.parse(data).usage.total_tokens.toString() + " tokens", {
					autoCloseDelay: 10000
				});
			});
		});
		// create request
		var request = identity.settings;
		request.prompt = prompt.replace(/^\/\w+\s+/, "").trim();
		request.prompt = identity.prepend + request.prompt + identity.append;
		req.write(m.JSON.stringify(request));
		req.end();
	}

	return {
		_init: _init,
		getAnswer: getAnswer
	};
};